local lgr = love.graphics
local win = love.window
local lfs = love.filesystem
local kbd = love.keyboard
local tmr = love.timer

bump = require 'bump.bump'

function love.load()
  font = {
    cga = lgr.newImageFont("cga.png", " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~")
  }

  smbTwo:load()
end

function love.update(dt)
  smbTwo:update(dt)
end

function love.draw()
  smbTwo:draw()

  lgr.draw(smbTwo.canvas, 0, 0, 0, 2, 2)
end

function love.keypressed(k, r)
  if k == 'escape' then love.event.quit() end
  smbTwo:keypressed(k, r)
end


smbTwo = {}

function smbTwo:load(arg)
  local timer = tmr.getTime()
  print("Loading Game State")
  local arg = arg or {}
  self.loaded = true
  self.backgroundUpdate = false
  self.backgroundDraw = true
  self.message = ""
  self.transition = 0
  self.name = "Test Bed"
  self.width = 640
  self.height = 360
  self.canvas = lgr.newCanvas(self.width, self.height)
  self.canvas:setFilter("linear", "nearest")

  self.map = {
    grid = {
      {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1},
      {1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,4,4,4,1},
      {1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,4,4,4,1},
      {1,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,3,3,1},
      {1,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,3,1},
      {1,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,0,9,9,9,9,9,9,1,0,0,0,0,0,0,0,1,0,0,1},
      {1,0,0,0,1,0,0,0,0,0,0,0,0,1,0,0,0,0,1,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,1,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,1,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,0,0,0,0,3,1},
      {1,0,0,0,0,0,0,0,0,0,0,2,0,0,0,1,0,0,0,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,2,0,0,0,1,0,0,2,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,2,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,2,0,0,0,0,0,0,2,0,0,1},
      {1,0,0,0,0,0,3,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1},
      {1,0,0,2,0,0,2,0,0,2,0,0,2,0,0,2,0,0,2,0,0,1},
      {1,0,0,2,0,0,2,0,0,2,0,0,2,0,0,2,0,0,2,0,0,1},
      {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1},
    },
    plantList = {
      { x = 10, y = 19, item = "bomb" },
      { x = 11, y = 19, item = "bomb" },
      { x = 12, y = 19, item = "bomb" },

      { x = 18, y = 19, item = "bomb" },

      { x = 20, y = 22, item = "veg" },
      { x = 21, y = 22, item = "veg" },
      { x = 22, y = 22, item = "veg" },
    },
    tileSize = 16
  }
  self.map.width = #self.map.grid
  self.map.height = #self.map.grid[1]
  print(self.map.width, self.map.height)

  self.map.collisionMap = bump.newWorld(128)

  local tempColMap = {}
  local markedGrid = {}
  self.map.plantGrid = {}
  for x = 1, self.map.width do
    tempColMap[x-1] = {}
    markedGrid[x-1] = {}
    self.map.plantGrid[x] = {}
    for y = 1, self.map.height do
      tempColMap[x-1][y-1] = self.map.grid[x][y]
      markedGrid[x-1][y-1] = false
      self.map.plantGrid[x][y] = { item = "" }
    end
  end

  for i = 1, #self.map.plantList do
    local p = self.map.plantList[i]
    self.map.plantGrid[p.x][p.y] = { item = p.item }
  end

  local boxSizeList = {
    {128,24}, {32,128}, {64,12}, {12,64}, {32,12}, {12,32}, {12,12}, {8,8}, {7,7}, {6,6}, {6,2}, {5,5}, {5,4}, {5,3},
    {5,2}, {4,4}, {2,4}, {4,5}, {4,2}, {4,3}, {4,2}, {3,3}, {3,4}, {1,32}, {32,1}, {30,1}, {20,1}, {1,20}, {7,1}, {6,1}, {5,1}, {4,1}, {2,3}, {1,4}, {3,1}, {1,3}, {2,2}, {2,1}, {1,2}, {1,1}
  }
  self.map.boxList = {}
  self.map.throwableBlockList = {}
  self.map.destroyableBlockList = {}
  self.projectiles = {}

  for currentTileType = 1, 9 do
    if currentTileType < 3 then
      for a = 1, #boxSizeList do
        local bw, bh = boxSizeList[a][1], boxSizeList[a][2]
        local count = bw * bh
        for x = 0, #tempColMap-bw+1 do
          for y = 0, #tempColMap[1]-bh+1 do
            local _count = 0
            for bx = x, x + (bw-1) do
              for by = y, y + (bh-1) do
                if not markedGrid[bx][by] then
                  if tempColMap[bx][by] == currentTileType then
                    _count = _count + 1
                  end
                end
              end
            end
            if _count == count then
              local n = #self.map.boxList+1
              local wallType = currentTileType
              self.map.boxList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = bw * self.map.tileSize, h = bh * self.map.tileSize, wallType = wallType, kind = "wall" }
              if wallType == 2 then
                self.map.boxList[n].solidTop = true
              end

              for bx = x, x + (bw-1) do
                for by = y, y + (bh-1) do
                  markedGrid[bx][by] = true
                end
              end
            end
          end
        end
      end
    elseif currentTileType == 3 then
      for x = 0, #tempColMap-1 do
        for y = 0, #tempColMap[1]-1 do
          if self.map.grid[x+1][y+1] == 3 then
            local n = "block" .. math.random(0,1000000)
            self.map.throwableBlockList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = 1 * self.map.tileSize, h = 1 * self.map.tileSize, wallType = 3, kind = "block", canPickUp = true, id = n }
          end
        end
      end
    elseif currentTileType == 4 then
      for x = 0, #tempColMap-1 do
        for y = 0, #tempColMap[1]-1 do
          if self.map.grid[x+1][y+1] == 4 then
            local n = "dWall_" .. x+1 .. "_" .. y+1
            self.map.destroyableBlockList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = 1 * self.map.tileSize, h = 1 * self.map.tileSize, wallType = 4, kind = "destroyableWall", id = n }
          end
        end
      end
    elseif currentTileType == 9 then
      for x = 0, #tempColMap-1 do
        for y = 0, #tempColMap[1]-1 do
          if self.map.grid[x+1][y+1] == 9 and self.map.grid[x+1][y] == 0 then
            local n = #self.map.boxList+1
            self.map.boxList[n] = { x = x * self.map.tileSize, y = y * self.map.tileSize, w = 1 * self.map.tileSize, h = 1 * self.map.tileSize, wallType = 2, kind = "wall" }
          end
        end
      end
    end
  end

  for i = 1, #self.map.boxList do
    self.map.collisionMap:add(self.map.boxList[i], self.map.boxList[i].x, self.map.boxList[i].y, self.map.boxList[i].w, self.map.boxList[i].h)
  end

  for i, b in pairs(self.map.throwableBlockList) do
    self.map.collisionMap:add(b, b.x, b.y, b.w, b.h)
  end

  for i, b in pairs(self.map.destroyableBlockList) do
    self.map.collisionMap:add(self.map.destroyableBlockList[i], b.x, b.y, b.w, b.h)
  end


  self.player = smbPlayer:create { x = 17 * self.map.tileSize, y = 20 * self.map.tileSize, w = math.floor(0.8 * self.map.tileSize), h = math.floor(1.3 * self.map.tileSize),
          maxFallSpeed = 10, jumping = false, jumpVelocity = 0, world = self.map.collisionMap, carryingBlockType = 0, carryingBlockID = 0, kind == "player" }

  self.map.collisionMap:add(self.player, self.player.x, self.player.y, self.player.w, self.player.h)

  win.setTitle("A loving tribute to Super Mario Bros. 2")

  -- lgr.setBackgroundColor(255,255,255)
end

function smbTwo:update(dt)
  local projectilesToRemove = {}

  self.player:update(dt)
  for i, p in pairs(self.projectiles) do
    p:update(dt)
    if p.isDead then
      projectilesToRemove[#projectilesToRemove+1] = i
    end
  end

  for i, p in pairs(projectilesToRemove) do
    -- self.map.collisionMap:remove(self.projectiles[p])
    self.projectiles[p]:destroy()
    self.projectiles[p] = nil
  end
end

function smbTwo:draw()
  self.canvas:clear(0,0,0,0)
  lgr.setCanvas(self.canvas)

  lgr.push()
  lgr.translate(-16,-16)
  for x = 1, self.map.width do
    for y = 1, self.map.height do
      if self.map.plantGrid[x][y].item ~= "" then
        lgr.setColor(0,200,0)
        lgr.rectangle("fill", x * 16 + 4, y * 16 - 4, 8, 4)
        -- lgr.rectangle("fill", x * 16, y * 16, 16, 16)
      end

      if self.map.grid[x][y] > 0 and self.map.grid[x][y] < 3 then
        lgr.push()
        lgr.translate(x * 16, y * 16)
        lgr.setColor(255,255,255)
        if self.map.grid[x][y] == 3 then
          lgr.setColor(100,200,255)
        end
        lgr.rectangle("fill", 0, 0, 16, 16)
        lgr.pop()
      end

      if self.map.grid[x][y] == 9 then
        lgr.push()
        lgr.translate(x * 16, y * 16)
        lgr.setColor(160,160,160)
        lgr.rectangle("fill", 1, 0, 2, 16)
        lgr.rectangle("fill", 2, 3, 12, 2)
        lgr.rectangle("fill", 2, 11, 12, 2)
        lgr.rectangle("fill", 13, 0, 2, 16)
        lgr.pop()
      end
    end
  end

  lgr.pop()

  for i = 1, #self.map.boxList do
    local b = self.map.boxList[i]
    if b.wallType == 1 then
      lgr.setColor(255,0,0,100)
    elseif b.wallType == 2 then
      lgr.setColor(0,255,0,100)
    else
      lgr.setColor(0,255,255,100)
    end
    lgr.rectangle("fill", b.x + 2, b.y + 2, b.w - 4, b.h - 4)
  end

  for i, b in pairs(self.map.throwableBlockList) do
    lgr.push()
    lgr.translate(0, 0)
    lgr.setColor(0,200,200)
    lgr.rectangle("fill", b.x, b.y, b.w, b.h / 2)
    lgr.setColor(0,255,255)
    lgr.rectangle("fill", b.x + 2, b.y + b.h / 2, b.w - 4, b.h / 2)
    lgr.pop()
  end

  for i, b in pairs(self.map.destroyableBlockList) do
    lgr.push()
    lgr.translate(0, 0)
    lgr.setColor(155,100,10)
    lgr.rectangle("fill", b.x, b.y, b.w, b.h)
    lgr.setColor(255,255,255,100)
    lgr.rectangle("fill", b.x, b.y, b.w-2, b.h-2)
    lgr.pop()
  end

  self.player:draw()

  for i, p in pairs(self.projectiles) do
    p:draw()
  end

  if kbd.isDown("lshift", "rshift") then
    lgr.setFont(font.cga)
    for x = 1, self.map.width do
      for y = 1, self.map.height do
        if self.map.grid[x][y] > 0 then
          lgr.push()
          lgr.translate((x-1) * 16 + 4, (y-1) * 16 + 4)
          lgr.setColor(0,0,0)
          lgr.print(self.map.grid[x][y], 1, 1)
          lgr.setColor(255,255,255)
          lgr.print(self.map.grid[x][y], 0, 0)
          lgr.pop()
        end
      end
    end
  end

  local _, n = self.map.collisionMap:check(self.player)
  local t = "Player:\n"
  t = t .. "Collisions: " .. n .. "\n"
  t = t .. "X: " .. mathRound(self.player.x, 4) .. "\n"
  t = t .. "Y: " .. mathRound(self.player.y, 4) .. "\n"
  t = t .. "VX: " .. self.player.vx .. "\n"
  t = t .. "VY: " .. self.player.vy .. "\n"
  t = t .. "TLT: " .. tostring(self.player.tlt) .. "\n"
  t = t .. "TLB: " .. tostring(self.player.tlb) .. "\n"
  t = t .. "TRT: " .. tostring(self.player.trt) .. "\n"
  t = t .. "TRB: " .. tostring(self.player.trb) .. "\n"
  t = t .. "Jumping: " .. tostring(self.player.isJumping) .. "\n"
  t = t .. "On Ground: " .. tostring(self.player.onGround) .. "\n"
  t = t .. "On Block: " .. tostring(self.player.standingOnBlock) .. "\n"
  t = t .. "X, Y: " .. tostring(self.player.pickBlockX) .. ", " .. tostring(self.player.pickBlockY) .. "\n"
  t = t .. "Carrying:\n     " .. tostring(self.player.carryingBlockType) .. "\n     " .. tostring(self.player.carryingBlockID) .. "\n"
  t = t .. "\nPress SHIFT to\nshow block grid\n\nArrows to move\nSPACE to pick up\nblock or veggie\n\nLadders do not\nwork yet!"

  lgr.setFont(font.cga)
  lgr.setColor(255,255,255)
  lgr.print(t, 500, 10)

  lgr.setCanvas()
end

function smbTwo:keypressed(k, r)
  if k == "p" then
    stateStack:replace(game, { saveFile = 2 })
  elseif k == " " then
    if self.player.carryingBlockType then
      -- Throw whatever is in your hand
      print(self.player.carryingBlockID, self.player.carryingBlockType)

      local vx = ((130 + math.abs(self.player.vx)) * self.player.facing)
      local vy = -(10 + math.abs(self.player.vy))
      if kbd.isDown("down") then
        vy = 30 + math.abs(self.player.vy)
      end
      local n = #self.projectiles+1
      local w, h = 16, 16
      if self.player.carryingBlockType == "block" then
        w = 16 - 4
        h = 16 - 8
      else
        w = 8
        h = 8
      end
      self.projectiles[n] = smbBlockProjectile:create {
        x = self.player.x - 2, y = (self.player.y + self.player.h) - 32, w = w, h = h, vx = vx, vy = vy,
        world = self.map.collisionMap, kind = self.player.carryingBlockType, id = self.player.carryingBlockID }
      self.map.collisionMap:add(self.projectiles[n], self.projectiles[n].x, self.projectiles[n].y, self.projectiles[n].w, self.projectiles[n].h)
      self.player.carryingBlockID = nil
      self.player.carryingBlockType = nil

      print("Hmm")
    elseif self.player.standingOnBlock == 3 and self.player.onGround then
      -- Check if standing on a pickupable block and pick it up
      self.map.collisionMap:remove(self.map.throwableBlockList[self.player.standingOnBlockID])
      self.map.throwableBlockList[self.player.standingOnBlockID] = nil
      self.map.grid[self.player.pickBlockX][self.player.pickBlockY] = 0
      self.player.carryingBlockID = self.player.standingOnBlockID
      self.player.carryingBlockType = "block"

      print("Picked up Block")
    elseif self.map.plantGrid[self.player.pickBlockX][self.player.pickBlockY].item ~= "" then
      -- Check for a plant in the ground and pull it up if available
      local p = self.map.plantGrid[self.player.pickBlockX][self.player.pickBlockY]
      print("Picked up a plant \"" .. p.item .. "\"")


      self.player.carryingBlockType = p.item

      p.item = ""
    else
      print("No")
    end
  end
end

function smbTwo:createExplosion(ix, iy)
  print("BOOM!", ix, iy)
  ix = (ix + 4) / self.map.tileSize
  iy = (iy + 4) / self.map.tileSize
  for x = ix - 1, ix + 1 do
    for y = iy - 1, iy + 1 do
      local nx, ny = math.floor(x + 0.5), math.floor(y + 0.5)
      if self.map.grid[nx] then
        if self.map.grid[nx][ny] then
          if self.map.grid[nx][ny] == 4 then
            self.map.collisionMap:remove(self.map.destroyableBlockList["dWall_"..nx.."_"..ny])
            self.map.destroyableBlockList["dWall_"..nx.."_"..ny] = nil
            self.map.grid[nx][ny] = 0
          end
        end
      end
    end
  end
end




smbPlayer = {}

function smbPlayer:create(v)
  local _i = {}
  setmetatable(_i, {__index = self})
  _i:setup(v or {})
  return _i
end
function smbPlayer:setup(v)
  self.world = v.world
  self.facing = 1
  self.x = v.x or 2
  self.y = v.y or 2
  self.w = v.w or width
  self.h = v.h or height
  self.vx = 0
  self.vy = 0
  self.isDead = false
  self.map = v.map or smbTwo.map
  self.gravityAccel  = 350 -- pixels per second^2
  self.deadDuration  = 3   -- seconds until res-pawn
  self.runAccel      = 300 -- the player acceleration while going left/right
  self.brakeAccel    = 500
  self.jumpVelocity  = 200 -- the initial upwards velocity when jumping
end

function smbPlayer:update(dt)
  self:changeVelocityByKeys(dt)
  self:changeVelocityByGravity(dt)
  self.vy = math.max(math.min(self.vy, 250), -250)
  self.vx = math.max(math.min(self.vx, 100), -100)

  if self.vy < 0 then
    self.movingVert = -1
  elseif self.vy > 0 then
    self.movingVert = 1
  else
    self.movingVert = 0
  end

  self:moveColliding(dt)
  self:changeVelocityByBeingOnGround(dt)


  -- self.tlt = self.map.grid[math.ceil((self.x) / self.map.tileSize)][math.ceil((self.y) / self.map.tileSize)]
  -- self.tlc = self.map.grid[math.ceil((self.x) / self.map.tileSize)][math.ceil((self.y + self.h / 2) / self.map.tileSize)]
  -- self.tlb = self.map.grid[math.ceil((self.x) / self.map.tileSize)][math.ceil((self.y + self.h) / self.map.tileSize)]
  -- self.trt = self.map.grid[math.ceil((self.x + self.w) / self.map.tileSize)][math.ceil((self.y) / self.map.tileSize)]
  -- self.trc = self.map.grid[math.ceil((self.x + self.w) / self.map.tileSize)][math.ceil((self.y + self.h / 2) / self.map.tileSize)]
  -- self.trb = self.map.grid[math.ceil((self.x + self.w) / self.map.tileSize)][math.ceil((self.y + self.h) / self.map.tileSize)]
  self.standingOnBlock = self.map.grid[math.ceil((self.x + (self.w / 2)) / self.map.tileSize)][math.ceil((self.y + self.h) / self.map.tileSize) + 1]
end

function smbPlayer:draw()
  lgr.setColor(255,100,255)
  lgr.rectangle("fill", math.floor(self.x), math.floor(self.y), self.w, self.h)
  lgr.setColor(0,0,255,200)
  lgr.rectangle("fill", math.floor(self.x + 1), math.floor(self.y + 1), self.w - 2, self.h - 2)

  if self.carryingBlockType then
    if self.carryingBlockType == "block" then
      lgr.push()
      lgr.translate(math.floor(self.x - 2), math.floor(self.y - 16))
      lgr.setColor(100,200,255)
      lgr.rectangle("fill", 0, 0, 16, 16 / 2)
      lgr.setColor(0,255,255)
      lgr.rectangle("fill", 0 + 2, 0 + 16 / 2, 16 - 4, 16 / 2)
      lgr.pop()
    else
      lgr.push()
      lgr.translate(math.floor(self.x + 2), math.floor(self.y - 8))
      lgr.setColor(100,200,255)
      lgr.rectangle("fill", 0, 0, 8, 8)
      lgr.pop()
    end
  end

end

function smbPlayer:changeVelocityByKeys(dt)
  if self.isDead then return end

  local vx, vy = self.vx, self.vy

  if love.keyboard.isDown("left") then
    vx = vx - dt * (vx > 0 and self.brakeAccel or self.runAccel)
    self.facing = -1
  elseif love.keyboard.isDown("right") then
    vx = vx + dt * (vx < 0 and self.brakeAccel or self.runAccel)
    self.facing = 1
  else
    local brake = dt * (vx < 0 and self.brakeAccel or -self.brakeAccel)
    if math.abs(brake) > math.abs(vx) then
      vx = 0
    else
      vx = vx + brake
    end
  end

  if love.keyboard.isDown("up") and self.onGround then -- jump/fly
    vy = -self.jumpVelocity
    self.isJumping = true
  end

  self.vx, self.vy = vx, vy
end

function smbPlayer:changeVelocityByBeingOnGround(dt)
  if self.onGround then
    self.vy = math.min(self.vy, 0)
  end
end

function smbPlayer:checkIfOnGround(ny)
  self.isJumping = false
  if ny < 0 then self.onGround = true end
end

local playerFilter = function(other)
  local p = smbTwo.player
  local okind = other.kind
  -- print(other.y, p.y + p.h, other.solidTop)
  -- if other.solidTop then
  --  if (p.movingVert < 1) then
  --    return false
  --  else
  --    -- if p.tlb == 2 or p.tlc == 2 or p.tlt == 2 or p.trb == 2 or p.trc == 2 or p.trt == 2 then
  --    --  return false
  --    -- end
  --    if p.y + p.h > other.y then
  --      return false
  --    end
  --    return true
  --  end
  -- end
  return (okind == "wall" or okind == "block" or okind == "destroyableWall" or not okind == "bomb") and not (other.solidTop and p.y + p.h > other.y)
end

function smbPlayer:moveColliding(dt)
  self.onGround = false
  local world = self.world

  local future_x = self.x + self.vx * dt
  local future_y = self.y + self.vy * dt

  local cols, len = world:check(self, future_x, future_y, playerFilter)
  if len == 0 then
    self:move(future_x, future_y)
  else
    local col, tl, tt, nx, ny, sl, st, other
    local visited = {}

    while len > 0 do
      col = cols[1]
      other = col.other
      if visited[other] then return end
      visited[other] = true

      tl,tt,nx,ny,sl,st = col:getSlide()

      self:changeVelocityByCollisionNormal(nx, ny)
      self:checkIfOnGround(ny)

      self:move(tl,tt)

      local tl, tt, nx, ny = col:getTouch()
      if ny == -1 then
        if self.standingOnBlock == 3 then
          self.standingOnBlockID = other.id or 0
        end
        self.pickBlockX = math.ceil((self.x + (self.w / 2)) / smbTwo.map.tileSize)
        self.pickBlockY = math.ceil((self.y + self.h) / smbTwo.map.tileSize) + 1
      end

      cols, len = world:check(self, sl, st, playerFilter)
      future_x, future_y = sl, st
    end

    self:move(future_x, future_y)
  end
end

function smbPlayer:changeVelocityByGravity(dt)
  self.vy = self.vy + self.gravityAccel * dt
end

function smbPlayer:changeVelocityByCollisionNormal(nx, ny, bounciness)
  bounciness = bounciness or 0
  local vx, vy = self.vx, self.vy

  if (nx < 0 and vx > 0) or (nx > 0 and vx < 0) then
    vx = -vx * bounciness
  end

  if (ny < 0 and vy > 0) or (ny > 0 and vy < 0) then
    vy = -vy * bounciness
  end

  self.vx, self.vy = vx, vy
end

function smbPlayer:move(l,t)
  self.x, self.y = l,t
  self.world:move(self, l, t)
end

function smbPlayer:getCenter()
  return self.x + self.w / 2,
         self.y + self.h / 2
end

function smbPlayer:destroy()
  self.world:remove(self)
end





smbBlockProjectile = {}

function smbBlockProjectile:create(v)
  local _i = {}
  setmetatable(_i, {__index = self})
  _i:setup(v or {})
  return _i
end
function smbBlockProjectile:setup(v)
  self.world = v.world
  self.facing = 1
  self.x = v.x or 2
  self.y = v.y or 2
  self.w = v.w or width
  self.h = v.h or height
  self.vx = v.vx or 100
  self.vy = v.vy or 0
  self.kind = v.kind or "block"
  self.map = v.map or smbTwo.map
  local g
  if self.kind == "block" then
    g = 350
  else
    g = 200
  end
  local solidTypes = { ["block"] = true, ["bomb"] = true, ["veg"] = false }
  self.solid = solidTypes[self.kind]

  if self.kind == "bomb" then
    self.timer = v.timer or 3
  end

  self.id = v.id or "projectile"..math.random(1,99999999)
  self.isDead = false
  self.gravityAccel  = g -- pixels per second^2
  self.deadDuration  = 3   -- seconds until res-pawn
  self.runAccel      = 300 -- the player acceleration while going left/right
  self.brakeAccel    = 500
  self.jumpVelocity  = 200 -- the initial upwards velocity when jumping

  self.timesBounced = 0
  self.maxBounceCount = v.maxBounce or 3
end

function smbBlockProjectile:update(dt)
  self:changeVelocityByGravity(dt)
  self.vy = math.max(math.min(self.vy, 250), -250)
  self.vx = math.max(math.min(self.vx, 100), -100)

  if self.timer then
    self.timer = self.timer - dt
    if self.timer <= 0 then
      self.timer = 0
      self.isDead = true
      print("DEAD")
      smbTwo:createExplosion(self.x + self.w / 2, self.y + self.h / 2)
    end
  end

  if self.solid then
    self:moveColliding(dt)
    self:changeVelocityByBeingOnGround(dt)
  else
    self:moveCollidingPassthrough(dt)
  end

  if self.isDead then
    if self.kind == "block" then
      local x = self.tx
      local y = self.ty
      self.map.grid[x+1][y+1] = 3

      local n = "block" .. math.random(0,1000000)
      self.map.throwableBlockList[n] = {
        x = x * self.map.tileSize, y = y * self.map.tileSize, w = 1 * self.map.tileSize, h = 1 * self.map.tileSize, wallType = 3, kind = "block", canPickUp = true, id = n }

      self.map.collisionMap:add(
        self.map.throwableBlockList[n], self.map.throwableBlockList[n].x, self.map.throwableBlockList[n].y, self.map.throwableBlockList[n].w, self.map.throwableBlockList[n].h)
    else

    end
  end
end

function smbBlockProjectile:draw()
  lgr.push()
  lgr.translate(math.floor(self.x), math.floor(self.y))
  if self.kind == "block" then
    lgr.push()
    lgr.translate(-2, -8)
    lgr.setColor(100,200,255)
    lgr.rectangle("fill", 0, 0, 16, 16 / 2)
    lgr.setColor(0,255,255)
    lgr.rectangle("fill", 0 + 2, 0 + 16 / 2, 16 - 4, 16 / 2)
    lgr.pop()
  else
    lgr.setColor(100,200,255)
    lgr.rectangle("fill", 0, 0, self.w, self.h)
  end
  lgr.setColor(255,0,0,200)
  if self.timer then
    if self.timer <= 1 then
      if math.cos(self.timer * 50) >= 0 then
        lgr.setColor(255,255,255,200)
      end
    end
  end
  lgr.rectangle("fill", 0, 0, self.w, self.h)
  if self.timer then
    lgr.setFont(font.tiny)
    lgr.setColor(0,0,0)
    lgr.printf(math.ceil(self.timer), 0, 1, 8, "center")
  end
  lgr.pop()
end

local blockProjectileFilter = function(other)
  if other.kind then
    return true
  end
  return false
end

function smbBlockProjectile:moveCollidingPassthrough(dt)
  self.onGround = false
  local world = self.world

  local future_x = self.x + self.vx * dt
  local future_y = self.y + self.vy * dt

  local cols, len = world:check(self, future_x, future_y, blockProjectileFilter)
  if len == 0 then
    self:move(future_x, future_y)
  else
    self:move(future_x, future_y)
  end

  if self.y > 500 then
    print("DEAD")
    self.isDead = true
  end
end

function smbBlockProjectile:changeVelocityByBeingOnGround(dt)
  if self.isDead then return end
  if self.onGround then
    self.vy = math.min(self.vy, 0)
    if self.vx > 0 then
      self.vx = self.vx - dt * 300
      if self.vx < 0 then self.vx = 0 end
    elseif self.vx < 0 then
      self.vx = self.vx + dt * 300
      if self.vx > 0 then self.vx = 0 end
    end

    -- print(self.vx, self.vy)
    if self.vx == 0 and self.vy == 0 then
      if self.kind == "block" then
        self.x = math.floor((self.x / 16) + 0.5) * 16
        self.isDead = true
        print("Landed")
      elseif self.kind == "bomb" then

      end

      self.tx = math.floor(self.x / 16)
      self.ty = math.floor(self.y / 16)
    end
  end
end

function smbBlockProjectile:checkIfOnGround(ny)
  self.isJumping = false
  if ny < 0 then self.onGround = true end
end

function smbBlockProjectile:moveColliding(dt)
  self.onGround = false
  local world = self.world

  local future_x = self.x + self.vx * dt
  local future_y = self.y + self.vy * dt

  local cols, len = world:check(self, future_x, future_y, blockProjectileFilter)
  if len == 0 then
    self:move(future_x, future_y)
  else
  local col, tl, tt, nx, ny, sl, st
  local visited = {}
  while len > 0 do
    col = cols[1]
    tl,tt,nx,ny,sl,st = col:getSlide()

    self:changeVelocityByCollisionNormal(nx, ny)
    self:checkIfOnGround(ny)

    self:move(tl,tt)

    if visited[col.other] then return end -- prevent infinite loops
      visited[col.other] = true

      local tl, tt, nx, ny = col:getTouch()
      if ny == -1 then
        if self.standingOnBlock == 3 then
          self.standingOnBlockID = col.other.id or 0
        end
        self.pickBlockX = math.ceil((self.x + (self.w / 2)) / smbTwo.map.tileSize)
        self.pickBlockY = math.ceil((self.y + self.h) / smbTwo.map.tileSize) + 1
      end

      cols, len = world:check(self, sl, st, blockProjectileFilter)
      if len == 0 then
        self:move(sl, st)
      end
      end
  end
end

function smbBlockProjectile:changeVelocityByGravity(dt)
  self.vy = self.vy + self.gravityAccel * dt
end

function smbBlockProjectile:changeVelocityByCollisionNormal(nx, ny, bounciness)
  bounciness = bounciness or 0.3
  local vx, vy = self.vx, self.vy

  if (nx < 0 and vx > 0) or (nx > 0 and vx < 0) then
    vx = -vx * bounciness
  end

  if (ny < 0 and vy > 0) or (ny > 0 and vy < 0) then
    vy = -vy * bounciness
    self.timesBounced = self.timesBounced + 1
    if self.timesBounced > self.maxBounceCount and vy < 0.01 then vy = 0 end
  end

  self.vx, self.vy = vx, vy
end

function smbBlockProjectile:move(l,t)
  self.x, self.y = l,t
  self.world:move(self, l, t)
end

function smbBlockProjectile:getCenter()
  return self.x + self.w / 2,
         self.y + self.h / 2
end

function smbBlockProjectile:destroy()
  self.world:remove(self)
end


smbDisposable = {}

function smbDisposable:create(v)
  local _i = {}
  setmetatable(_i, {__index = self})
  _i:setup(v or {})
  return _i
end
function smbDisposable:setup(v)
  self.life = v.life or 1
  self.x = v.x or 0
  self.y = v.y or 0
  self.isDead = false
end

function smbDisposable:update(dt)
  if self.isDead then return false end
  self.life = self.life - dt
  if self.life < 0 then
    self.life = 0
    self.isDead = true
  end
end

function smbDisposable:draw()

end



function mathRound(val, decimal)
  if not val then return 0 end
  if (decimal) then
    return math.floor( (val * 10^decimal) + 0.5) / (10^decimal)
  else
    return math.floor(val+0.5)
  end
end

